#ifndef MAINWIDGET_H
#define MAINWIDGET_H

#include <QObject>
#include <QtOpenGL>
#include <QOpenGLWidget>
#include <QOpenGLFunctions_4_3_Core>
#include <QOpenGLShaderProgram>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include "mesh.h"
#include "camera.h"


class MainWidget : public QOpenGLWidget, public QOpenGLFunctions_4_3_Core
{
    Q_OBJECT
public:
    MainWidget(Camera *cam);

    QSize minimumSizeHint() const;
    QSize sizeHint() const;

public slots:
    void loadNewMesh();
    void setInterpolation(bool linear);
    void setRotation(bool started);
    void setAnimating(bool started);
    void nextAnimation();


protected:
        void initializeGL();
        void paintGL();
        void resizeGL(int width, int height);

        void mousePressEvent(QMouseEvent *event ) ;
        void mouseMoveEvent(QMouseEvent *event ) ;
        void wheelEvent(QWheelEvent *event ) ;

private:
        QVector3D* MainWidget::trackballPoint(int x, int y);
        void rotate(QVector3D *newPos );
        void move(QPointF *newPos );

        Camera *cam;

        QOpenGLShaderProgram* animationShader;
        QMatrix4x4 m_projection;
        QTime startTime;
        QTime rotTime;

        float camDistance;

        QVector3D *lastSpeherePos;
        QPointF * lastScreenPos;


        boolean linear;
        boolean rotation;
        boolean animate;
        int Animation;


        Mesh *mesh;

        QOpenGLShaderProgram* initShader(QString fragSource, QString vertSource);
};

#endif // MAINWIDGET_H
