#include "mesh.h"

void Vertex::AddBoneData(uint BoneID, float Weight){

    for (int i = 0 ; i < 4 ; i++) {
        if (weights[i] == 0.0) {
            IDs[i]     = BoneID;
            weights[i] = Weight;
            //qDebug()<<"Add"<<i<<BoneID<<Weight<<pos;
            return;
        }
    }
    qCritical()<<"More tahn 4 Bones at Vertex"<<pos<<BoneID<<Weight<<"added:"
              <<IDs[0]<<weights[0]<<IDs[1]<<weights[1]<<IDs[2]<<weights[2]<<IDs[3]<<weights[3];

}

Mesh::MeshEntry::MeshEntry()
{
    materialIndex = 0xFFFFFFFF;
    numIndex = 0;
    VB = 0xFFFFFFFF;
    IB = 0xFFFFFFFF;
    double amax = std::numeric_limits<float>::max();
    min = QVector3D(amax,amax,amax);
    max = QVector3D(-amax,-amax,-amax);
}

void Mesh::MeshEntry::init(QOpenGLFunctions_4_3_Core *f,QVector<Vertex>& Vertices,
                           QVector<unsigned int>& Indices){
    this->f = f;
    numIndex = Indices.size();

    f->glGenBuffers(1, &VB);
    f->glBindBuffer(GL_ARRAY_BUFFER, VB);
    f->glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * Vertices.size(), &Vertices[0], GL_STATIC_DRAW);

    f->glGenBuffers(1, &IB);
    f->glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, IB);
    f->glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int) * numIndex, &Indices[0], GL_STATIC_DRAW);

    //calc AABB
    for (int i = 0; i < Vertices.size(); ++i) {
        Vertex v = Vertices[i];
        min.setX(qMin(min.x(),v.pos.x()));
        min.setY(qMin(min.y(),v.pos.y()));
        min.setZ(qMin(min.z(),v.pos.z()));

        max.setX(qMax(max.x(),v.pos.x()));
        max.setY(qMax(max.y(),v.pos.y()));
        max.setZ(qMax(max.z(),v.pos.z()));
    }
}

Mesh::MeshEntry::~MeshEntry()
{
    f->glDeleteBuffers(1, &VB);
    f->glDeleteBuffers(1, &IB);
}

Mesh::MaterialInfo::MaterialInfo()
{
    Name = QString("Default");
    Diffuse = QVector3D();
    Specular= QVector3D();
    hasTexture = false;
    texture = Texture();
    Shininess = 30;
}

Mesh::Node::Node()
{
    name = QString("DefaultNode");
    transformation.setToIdentity();
    meshes.clear();
    children.clear();
}

Mesh::Mesh(QOpenGLFunctions_4_3_Core *f,QString fileName)
{
    loaded = false;
    linear = false;
    this->f =f;
    numberOfBones = 0;
    scene = NULL;
    scene= importer.ReadFile(fileName.toStdString(),
                             //aiProcess_CalcTangentSpace |
                             aiProcess_GenSmoothNormals |
                             //aiProcess_JoinIdenticalVertices |
                             //aiProcess_SortByPType |
                             // ? aiProcess_FlipUVs |
                             aiProcess_Triangulate
                             );

    qDebug()<<"File Read";
    if( !scene)
    {
        qDebug()<<importer.GetErrorString();
    }else {
        QString dir = QFileInfo(fileName).path();

        //Init Materials
        materials.resize(scene->mNumMaterials);
        if(scene->HasMaterials()){
            //Init Materials
            for (unsigned int i = 0 ; i < scene->mNumMaterials ; i++) {
                const aiMaterial* material = scene->mMaterials[i];
                aiString mname;
                material->Get(AI_MATKEY_NAME, mname);
                initMaterial(dir, i, material);

                qDebug()<<"Loaded Material"<<(i+1)<<"/"<<scene->mNumMaterials<<":"<<mname.C_Str()<<"Shininess"<<materials[i].Shininess;
            }
        }
        qDebug()<<"Materials Read";

        //Init Mesh
        if(scene->HasMeshes())
        {
            entries.resize(scene->mNumMeshes);
            for (int i = 0 ; i < entries.size() ; i++) {
                aiMesh* meshEntry = scene->mMeshes[i];
                initMeshEntry(i, meshEntry);

            }
        } else{
            qWarning()<<"No Mesh found";
            return;
        }

        if(scene->HasLights()){
            qDebug()<<"Ligths found";
        }
        qDebug()<<"Mesh Read";

        //init Nodes
        if(scene->mRootNode != NULL)
        {
            initNode(scene, scene->mRootNode, rootNode, QString());
        } else {
            rootNode.transformation.setToIdentity();
            rootNode.children.resize(0);
            rootNode.meshes.resize(entries.length());

            for(int i = 0; i < entries.length(); ++i)
            {
                rootNode.meshes[i] = i;
            }
            qDebug()<<"No Root Node";
        }
        qDebug()<<"Nodes Read";

        globalInverseTransform = rootNode.transformation.inverted();

        double amax = std::numeric_limits<float>::max();
        QVector3D min = QVector3D(amax,amax,amax);
        QVector3D max = QVector3D(-amax,-amax,-amax);

        findObjectDimension(rootNode,QMatrix4x4(),min,max);
        qDebug()<<"AABB"<<min<<max;

        float dist = qMax(max.x() - min.x(), qMax(max.y()-min.y(), max.z() - min.z()));
        float sc = 100.0/dist;
        QVector3D center = (max - min)/2;
        QVector3D trans = -(max - center);

        // Apply the scale and translation to a matrix
        screenTransform.setToIdentity();
        screenTransform.scale(sc);
        screenTransform.translate(trans);

        loaded = true;
    }
}

Mesh::~Mesh()
{
    entries.clear();
    materials.clear();
}

void Mesh::initNode(const aiScene *scene, aiNode *node, Node &newNode,QString debugoffset){
    newNode.name = node->mName.length != 0 ? node->mName.C_Str() : "";
    newNode.transformation = QMatrix4x4(node->mTransformation[0]);

    newNode.meshes.resize(node->mNumMeshes);
    bool debug = false;
    if(debug) qDebug()<<debugoffset.toStdString().c_str() << "NodeName" << newNode.name;
    if(boneMap.find(newNode.name) != boneMap.end()){
        if(debug)  qDebug()<<debugoffset.toStdString().c_str() << "hasBone";
    }

    if(debug) qDebug()<<debugoffset.toStdString().c_str() << "  NumMeshes" << newNode.meshes.size();

    for(uint i = 0; i < node->mNumMeshes; ++i)
    {
        newNode.meshes[i] = node->mMeshes[i];
        if(debug) qDebug()<<debugoffset.toStdString().c_str() << "    MeshName" << entries[newNode.meshes[i]].name;
    }

    if(debug) qDebug()<<debugoffset.toStdString().c_str() << "  NumChildren" << node->mNumChildren;
    QString newDebug = QString(debugoffset).append("    ");
    for(uint i = 0; i < node->mNumChildren; ++i)
    {
        newNode.children.push_back(Node());
        initNode(scene, node->mChildren[i], newNode.children[i],newDebug);
    }

}

void Mesh::initMaterial(QString dir, unsigned int i, const aiMaterial* material)
{
    aiString mname;
    material->Get(AI_MATKEY_NAME, mname);
    if (mname.length > 0)
        materials[i].Name = QString(mname.C_Str());

    aiColor3D dif(0.f,0.f,0.f);
    aiColor3D amb(0.f,0.f,0.f);
    aiColor3D spec(0.f,0.f,0.f);
    float shine = 0.0;

    material->Get(AI_MATKEY_COLOR_AMBIENT, amb);
    material->Get(AI_MATKEY_COLOR_DIFFUSE, dif);
    material->Get(AI_MATKEY_COLOR_SPECULAR, spec);
    material->Get(AI_MATKEY_SHININESS, shine);

    materials[i].Diffuse = QVector3D(dif.r, dif.g, dif.b);
    materials[i].Specular = QVector3D(spec.r, spec.g, spec.b);
    materials[i].Shininess = shine;

    if (materials[i].Shininess == 0.0){
        materials[i].Shininess = 2;

    }


    if (material->GetTextureCount(aiTextureType_DIFFUSE) > 0) {
        aiString Path;

        if (material->GetTexture(aiTextureType_DIFFUSE, 0, &Path, NULL, NULL, NULL, NULL, NULL) == AI_SUCCESS) {
            QString temp = QString(Path.data);
            temp = temp.split("\\").last();
            temp = temp.split("/").last();
            std::string FullPath = dir.toStdString() + "/" + temp.toStdString();
            materials[i].texture.Load(GL_TEXTURE_2D, QString(FullPath.c_str()));
            materials[i].hasTexture = true;
        } else{
            qDebug()<<"Warning No Texture";
            materials[i].hasTexture = false;
        }

    }
}

void Mesh::initMeshEntry(int index, aiMesh * entry){

    QVector<Vertex> Vertices;
    QVector<unsigned int> Indices;

    const aiVector3D Zero3D(0.0f, 0.0f, 0.0f);

    for (unsigned int i = 0 ; i < entry->mNumVertices ; i++) {
        const aiVector3D* pPos      = &(entry->mVertices[i]);
        const aiVector3D* pNormal   = &(entry->mNormals[i]);
        const aiVector3D* pTexCoord = entry->HasTextureCoords(0) ? &(entry->mTextureCoords[0][i]) : &Zero3D;

        //possible: nuuv Channels, TAngents and Bitangents

        Vertex v(QVector3D(pPos->x, pPos->y, pPos->z),
                 QVector3D(pNormal->x, pNormal->y, pNormal->z),
                 QVector2D(pTexCoord->x, pTexCoord->y));

        Vertices.push_back(v);
    }

    for (unsigned int i = 0 ; i < entry->mNumFaces ; i++) {
        const aiFace& face = entry->mFaces[i];
        assert(face.mNumIndices == 3);
        Indices.push_back(face.mIndices[0]);
        Indices.push_back(face.mIndices[1]);
        Indices.push_back(face.mIndices[2]);
    }


    entries[index].name = entry->mName.length != 0 ? entry->mName.C_Str() : "";
    entries[index].materialIndex = entry->mMaterialIndex;

    qDebug()<<"Loaded Mesh:"<<entries[index].name<<"HasBones:"<<entry->HasBones();

    if(entry->HasBones()){
        for(uint i = 0; i< entry->mNumBones; i++){
            int index = 0;
            QString BoneName = QString(entry->mBones[i]->mName.data);

            if (boneMap.find(BoneName) == boneMap.end()) {
                // Allocate an index for a new bone
                index = numberOfBones;
                numberOfBones++;
                BoneInfo bi;
                bones.push_back(bi);
                bones[index].BoneOffset = QMatrix4x4(entry->mBones[i]->mOffsetMatrix[0]);
                boneMap[BoneName] = index;
            }
            else {
                index = boneMap[BoneName];
            }

            for (uint j = 0 ; j < entry->mBones[i]->mNumWeights ; j++) {
                uint VertexID = entry->mBones[i]->mWeights[j].mVertexId;
                float Weight  = entry->mBones[i]->mWeights[j].mWeight;
                Vertices[VertexID].AddBoneData(index, Weight);
            }

        }
    }

    entries[index].init(f,Vertices, Indices);

}

void Mesh::render(QOpenGLShaderProgram *shader, QMatrix4x4 V,QMatrix4x4 P, float TimeInSeconds, int animation){
    if(!loaded)
        return;
    if(f == NULL){
        qDebug()<<"f = null";
        return;
    }

    f->glEnableVertexAttribArray(positionIndex);
    f->glEnableVertexAttribArray(normalIndex);
    f->glEnableVertexAttribArray(uvIndex);
    f->glEnableVertexAttribArray(boneIndex);
    f->glEnableVertexAttribArray(boneweightIndex);


    updateBoneTransform(TimeInSeconds,animation);
    QVector<QMatrix4x4> boneTransforms = QVector<QMatrix4x4>(bones.size());
    for(int i = 0; i < bones.size(); i++){
        boneTransforms[i] = bones[i].FinalTransformation;
    }
    if(boneTransforms.size() == 0){
        boneTransforms.push_back(QMatrix4x4());
    }

    shader->setUniformValueArray("Bones",boneTransforms.constData(),boneTransforms.size());

    renderNode(shader,rootNode,V*screenTransform,P,QMatrix4x4());


    f->glDisableVertexAttribArray(positionIndex);
    f->glDisableVertexAttribArray(normalIndex);
    f->glDisableVertexAttribArray(uvIndex);
    f->glDisableVertexAttribArray(boneIndex);
    f->glDisableVertexAttribArray(boneweightIndex);



}

void Mesh::renderNode(QOpenGLShaderProgram *shader, Node &node, QMatrix4x4 V,QMatrix4x4 P,QMatrix4x4 M){
    M *= node.transformation;

    QMatrix4x4 MV = V*M;
    QMatrix4x4 MVP = P*MV;
    QMatrix3x3 normalMat = MV.normalMatrix();

    shader->setUniformValue("MVP",MVP);
    shader->setUniformValue("MV",MV);
    shader->setUniformValue("N",normalMat);

    for (int i = 0 ; i < node.meshes.size() ; i++) {
        int index = node.meshes[i];
        //load Material
        renderMesh(shader, index);
    }

    for (int i = 0 ; i < node.children.size() ; i++) {
        renderNode(shader,node.children[i],V,P,M);
    }
}

void Mesh::renderMesh(QOpenGLShaderProgram *shader, int index)
{
    int MaterialIndex = entries[index].materialIndex;
    if (MaterialIndex < materials.size()) {
        shader->setUniformValue("materialInfo.Diffuse",materials[MaterialIndex].Diffuse);
        shader->setUniformValue("materialInfo.Specular",materials[MaterialIndex].Specular);
        shader->setUniformValue("materialInfo.Shininess",materials[MaterialIndex].Shininess);
        shader->setUniformValue("materialInfo.hasTexture",materials[MaterialIndex].hasTexture);

        if(materials[MaterialIndex].hasTexture)
            materials[MaterialIndex].texture.bind(f,GL_TEXTURE0);
    }


    // Draw Vertex Array
    f->glBindBuffer(GL_ARRAY_BUFFER, entries[index].VB);
    f->glVertexAttribPointer(positionIndex, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), 0);
    f->glVertexAttribPointer(normalIndex, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)12); //12
    f->glVertexAttribPointer(uvIndex, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)24); // 12+12
    f->glVertexAttribIPointer(boneIndex, 4, GL_INT, sizeof(Vertex), (const GLvoid*)32); // 12+12+8
    f->glVertexAttribPointer(boneweightIndex, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)48); //12+12+8+16

    f->glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, entries[index].IB);
    f->glDrawElements(GL_TRIANGLES, entries[index].numIndex, GL_UNSIGNED_INT, 0);
}

//Bone Update
void Mesh::updateBoneTransform(float TimeInSeconds, int animIndex){
    if(!scene || bones.size() == 0|| !scene->HasAnimations())
        return;
    animIndex = animIndex % scene->mNumAnimations;

    const aiAnimation* animation = scene->mAnimations[animIndex];


    float TicksPerSecond = (float)(animation->mTicksPerSecond != 0 ? animation->mTicksPerSecond : 25.0f);
    float TimeInTicks = TimeInSeconds * TicksPerSecond;
    float AnimationTime = fmod(TimeInTicks, (float)animation->mDuration);

    updateBoneHeirachy(AnimationTime,animation,this->rootNode,QMatrix4x4());



}

void Mesh::updateBoneHeirachy(float AnimTime,const aiAnimation* animation, Node node, QMatrix4x4 parentTransform){

    QMatrix4x4 nodeTransformation = node.transformation;


    //Find channel
    const aiNodeAnim *nodeAnimation = findAnimOfNode(animation,node.name);

    //calc Tcurrent Trasnformation;
    if (nodeAnimation) {
        QVector3D scale = calcInterpolatedScaling(AnimTime,nodeAnimation);
        QQuaternion rot = calcInterpolatedRotation(AnimTime,nodeAnimation);
        QVector3D trans = calcInterpolatedTranslation(AnimTime,nodeAnimation);
        //qDebug()<<"calc"<<scale<<rot<<trans;

        nodeTransformation.setToIdentity();
        nodeTransformation.translate(trans);
        nodeTransformation.rotate(rot);
        nodeTransformation.scale(scale);
    }


    //Update Bone and children
    QMatrix4x4 GlobalTransform = parentTransform * nodeTransformation;

    if (boneMap.find(node.name) != boneMap.end()) {
        int index = boneMap[node.name];
        bones[index].FinalTransformation = globalInverseTransform * GlobalTransform * bones[index].BoneOffset;
    }

    for(int i = 0; i < node.children.size(); ++i)
    {
        updateBoneHeirachy(AnimTime,animation,node.children[i],GlobalTransform);
    }

}

const aiNodeAnim* Mesh::findAnimOfNode(const aiAnimation* animation, QString NodeName){
    for (uint i = 0 ; i < animation->mNumChannels ; i++) {
        const aiNodeAnim* nodeanim = animation->mChannels[i];

        if (QString(nodeanim->mNodeName.data) == NodeName) {
            return nodeanim;
        }
    }

    return NULL;
}

//Interpolarisation
QQuaternion Mesh::calcInterpolatedRotation(float AnimTime,const aiNodeAnim *nodeAnimation){
    if(nodeAnimation->mNumRotationKeys == 1){
        return convert(nodeAnimation->mRotationKeys[0].mValue);
    }

    //Index find
    int index = 0;
    for (uint i = 0 ; i < nodeAnimation->mNumRotationKeys - 1 ; i++) {
        if (AnimTime < (float)nodeAnimation->mRotationKeys[i + 1].mTime) {
            index = i;
            break;
        }
    }

    float DeltaTime = (float)(nodeAnimation->mRotationKeys[index+1].mTime - nodeAnimation->mRotationKeys[index].mTime);
    float Factor = (AnimTime - (float)nodeAnimation->mRotationKeys[index].mTime) / DeltaTime;
    // qDebug()<<AnimTime<<DeltaTime<<Factor;

    const aiQuaternion start = nodeAnimation->mRotationKeys[index].mValue;
    const aiQuaternion end = nodeAnimation->mRotationKeys[index+1].mValue;

    aiQuaternion RotationQ;
    aiQuaternion::Interpolate(RotationQ,start,end,Factor);
    return convert(RotationQ);
    //return convert(start);
}

QVector3D Mesh::calcInterpolatedScaling(float AnimTime,const aiNodeAnim *nodeAnimation){
    return calcInterpolatedVectorKey(AnimTime, nodeAnimation,nodeAnimation->mScalingKeys,nodeAnimation->mNumScalingKeys);
}

QVector3D Mesh::calcInterpolatedTranslation(float AnimTime, const aiNodeAnim *nodeAnimation){
    return calcInterpolatedVectorKey(AnimTime,nodeAnimation, nodeAnimation->mPositionKeys,nodeAnimation->mNumPositionKeys);
}

QVector3D Mesh::calcInterpolatedVectorKey(float AnimTime,const aiNodeAnim *nodeAnimation, const aiVectorKey* keys, const int numberOfKeys){
    if(numberOfKeys == 1){
        return convert(keys[0].mValue);
    }

    //Index find
    int index = 0;
    for (int i = 0 ; i < numberOfKeys - 1; i++) {
        if (AnimTime < (float)keys[i + 1].mTime) {
            index = i;
            break;
        }
    }

    float DeltaTime = (float)(keys[index+1].mTime - keys[index].mTime);
    float Factor = (AnimTime - (float)keys[index].mTime) / DeltaTime;

    QVector3D start = convert(keys[index].mValue);
    QVector3D end = convert(keys[index+1].mValue);



    if(start.distanceToPoint(end) < EPSILON){ // same keyframe
        return start;
    }

    if(linear){
        return linInterpolate(start,end,Factor);
    }

    if(index == 0 ){ //Start
        switch (nodeAnimation->mPreState) {
        case aiAnimBehaviour_CONSTANT: // nearest Key
            return Factor <0.5f ? start : end;
            break;
        case aiAnimBehaviour_REPEAT:// use start or end
        {
            QVector3D before = convert(keys[index-1].mValue);
            QVector3D after = convert(keys[numberOfKeys - 1].mValue);
            return Catmul(before,start,end,after,Factor);
        }
        case aiAnimBehaviour_LINEAR: // linear interpolate
            return linInterpolate(start,end,Factor);
        case aiAnimBehaviour_DEFAULT: // take default node transformation
            //TODO aiNodes haben nur translation rotation und scalierung nicht getrennt gespeichert
        default:
            return linInterpolate(start,end,Factor);
        }
    } else if((index+2)== numberOfKeys){ // Ende
        switch (nodeAnimation->mPostState) {
        case aiAnimBehaviour_CONSTANT: // nearest Key
            return Factor <0.5f ? start : end;
            break;
        case aiAnimBehaviour_REPEAT: // use start or end
        {
            QVector3D before = convert(keys[index-1].mValue);
            QVector3D after = convert(keys[0].mValue);
            return Catmul(before,start,end,after,Factor);
        }
        case aiAnimBehaviour_LINEAR: // linear interpolate
            return linInterpolate(start,end,Factor);
        case aiAnimBehaviour_DEFAULT: // take default node transformation
            //TODO
        default:
            return linInterpolate(start,end,Factor);
        }
    } else{
        QVector3D before = convert(keys[index-1].mValue);
        QVector3D after = convert(keys[index+2].mValue);

        if(before.distanceToPoint(start) < EPSILON
                ||end.distanceToPoint(after)<EPSILON){ //  keyframes before or after are the same. No Nomal computation
            //TODO
            return linInterpolate(start,end,Factor);
        }




        return Catmul(before,start,end,after,Factor);
    }

}

QVector3D Mesh::Catmul(QVector3D P0, QVector3D P1, QVector3D P2, QVector3D P3,float Factor){
    double t0 = 0;
    double t1 = timeCatmul(P0,P1,t0);
    double t2 = timeCatmul(P1,P2,t1);
    double t3 = timeCatmul(P2,P3,t2);
    double t = Factor;
    t = (1-Factor)*t1+Factor*t2;
    //qDebug()<<before<<start<<end<<after;
    //qDebug()<<t0<<t1<<t2<<t3<<t;

    QVector3D A1 = (t1-t)/(t1-t0)*P0    + (t-t0)/(t1-t0)*P1;
    QVector3D A2 = (t2-t)/(t2-t1)*P1    + (t-t1)/(t2-t1)*P2;
    QVector3D A3 = (t3-t)/(t3-t2)*P2    + (t-t2)/(t3-t2)*P3;

    QVector3D B1 = (t2-t)/(t2-t0)*A1 + (t-t0)/(t2-t0)*A2;
    QVector3D B2 = (t3-t)/(t3-t1)*A2 + (t-t1)/(t3-t1)*A3;

    QVector3D C  = (t2-t)/(t2-t1)*B1 + (t-t1)/(t2-t1)*B2;

    if(C.isNull() || qIsNaN(C.x())){
        qCritical()<<"Catmul bad:"<<C<<"from"<<P0<<P1<<P2<<P3;
    }

    return C;


}

double Mesh::timeCatmul(QVector3D p1, QVector3D p2, double t_1){
    double dist = p1.distanceToPoint(p2);
    double alpha = 0.0; //uniform
    alpha = 0.5; //centripetal
    //float alpha = 1.0; //chordal

    double t =  qPow(dist,alpha) +t_1;
    // qDebug()<<dist<<t;
    return t;
}

void Mesh::findObjectDimension(Node node, QMatrix4x4 transform, QVector3D &min, QVector3D &max){
    transform *= node.transformation;

    for (int i = 0; i < node.meshes.size(); i++) {

        QVector4D temp = transform*QVector4D(entries[node.meshes[i]].min,1.0);
        min.setX(qMin(min.x(),temp.x()));
        min.setY(qMin(min.y(),temp.y()));
        min.setZ(qMin(min.z(),temp.z()));
        max.setX(qMax(max.x(),temp.x()));
        max.setY(qMax(max.y(),temp.y()));
        max.setZ(qMax(max.z(),temp.z()));

        temp = transform*QVector4D(entries[node.meshes[i]].max,1.0);
        min.setX(qMin(min.x(),temp.x()));
        min.setY(qMin(min.y(),temp.y()));
        min.setZ(qMin(min.z(),temp.z()));
        max.setX(qMax(max.x(),temp.x()));
        max.setY(qMax(max.y(),temp.y()));
        max.setZ(qMax(max.z(),temp.z()));
    }

    for (int i = 0; i < node.children.size(); i++) {
        findObjectDimension(node.children[i], transform, min, max);
    }
}



